'''
Stateful, singleton, command and control center. This is the brains of the three 
paradrop components: tools, router, and server.

Takes over the functionality of the following modules, globals, and all around cruft:
    settings
    storage (keys, settings, names, permissions)
    chuteStore
    port/host consts
    portal and server administration

Additionally:
    Persists data to disk synchronously (for now, unless it gets hairy.)
    Manages init and deinit
    Document versioning, migrations, and validations through barit.

This object is meant to be subclassed by each of the three PD components. 
They should build upon the methods below as it suits their needs-- for example, 
tools may need to store information about routers, routers will store 
chutes, and the server will rely heavily on riffle callbacks.

Should make this a singleton. And make a print thread for this guy, just for safety.

The base class does not implement any riffle integration other than 
setting up the keys and default connection information. You should register for 
connection callbacks using smokesignals (see paradrop/main.py). DO NOT
store connections yourself-- riffle already does that and you'll screw up the portal!
'''

import os
import yaml

from twisted.internet import reactor
import smokesignal

from pdtools.lib import store, riffle, output

# Global access. Assign this wherever you instantiate the Nexus object:
#       nexus.core = MyNexusSubclass()
core = None


class NexusBase(object):

    def __init__(self, nexType, mode='development', stealStdio=True, printToConsole=True):
        '''
        The one big thing this function leaves out is reactor.start(). Call this externally 
        *after* initializing a nexus object. 

        :param type: one of [router, tool, server]
        :type type: str.
        :param mode: one of [production, development, test, local]. Changes host and ports.
            See section "network resolution" below 
        '''

        self.type, self.mode = nexType, mode

        if mode not in 'production development test local'.split():
            raise KeyError("Mode " + self.mode + " not valid")

        # initialize the paths and load existing settings
        self.makePaths()
        self.load()

        # initialize output. If filepath is set, logs to file.
        # If stealStdio is set intercepts all stderr and stdout and interprets it internally
        # If printToConsole is set (defaults True) all final output is rendered to stdout
        output.out.startLogging(filePath=self.logPath, stealStdio=stealStdio, printToConsole=printToConsole)

        # Asssign global riffle keys
        # riffle.portal.keyPrivate = self.getKey('pub')
        # riffle.portal.certCa = self.getKey('ca')
        # riffle.portal.host = self.serverHost()
        # riffle.portal.port = self.rifflePort()

        # register onStop for the shutdown call
        reactor.addSystemEventTrigger('before', 'shutdown', self.onStop)
        reactor.callLater(0, self.onStart)

    def onStart(self):
        pdid = self.get('pdid') if self.provisioned() else 'unprovisioned router'

        output.out.usage('%s (%s) coming up' % (pdid, self.type))

    def onStop(self):
        self.save()

        # riffle.portal.close()

        # remove any and all pubsub registrations. If not done, this can cause
        # issues with subs that have to do with network connections

        output.out.usage('%s going down' % self.type)
        smokesignal.clear_all()
        output.out.endLogging()

    def makePaths(self):
        '''
        Are we on a VM? On snappy? Bare metal? The server?  So many paths, so few answers!
        '''

        # Lets start out being a router
        self.rootPath = os.getenv("SNAP_APP_USER_DATA_PATH", None)

        # Put tools contents in the home directory
        if self.type is 'tool':
            self.rootPath = os.path.expanduser('~') + '/.paradrop/'

        # Current directory (since we might have more than one server at a time)
        elif self.type is 'server':
            self.rootPath = os.getcwd() + '/.paradrop/'

        # We're a router, yay! But are we a router on snappy or local? The snappy
        # path will not resolve on a local machine-- let it fall through, we're set
        elif not self.rootPath:
            self.rootPath = os.path.expanduser('~') + '/.paradrop/local/'

        # Set the boring paths
        self.logPath = self.rootPath + '/logs/'
        self.keyPath = self.rootPath + '/keys/'
        self.miscPath = self.rootPath + '/misc/'
        self.configPath = self.rootPath + '/config'  # this is a path, not a file

        # create the paths
        for x in [self.rootPath, self.logPath, self.keyPath, self.miscPath]:
            if not os.path.exists(x):
                os.makedirs(x)

    def load(self):
        '''
        Load our settings/configuration data from the path ivars. Does not check if the paths are None.

        This is a combo of functionality from store and settings. The method of loading is configurable. 
        Right now its YAML.
        '''

        # Check to make sure we have a default settings file
        if not os.path.isfile(self.configPath):
            createDefaultInfo(self.configPath)

        self.config = loadYaml(self.configPath)

        # Sanity check contents of info and throw it out if bad
        if not validateInfo(self.config):
            out.err('Saved configuration data invalid, destroying it.')
            os.remove(self.configPath)
            createDefaultInfo(self.configPath)
            self.config = loadYaml(self.configPath)

    def save(self):
        ''' Ehh. Ideally this should happen constantly and asynchronously. '''
        writeYaml(self.config, self.configPath)

    #########################################################
    # Settings
    #########################################################

    def set(self, k, v):
        self.config[k] = v
        self.save()

    def get(self, k):
        return self.config[k]

    def provisioned(self):
        return self.config['pdid'] is not None

    #########################################################
    # Keys
    #########################################################

    def saveKey(self, key, name):
        ''' Save the key with the given name. Overwrites by default '''
        path = self.keyPath + name

        with open(path, 'wb') as f:
            f.write(key)

    def getKey(self, name):
        ''' Returns the given key or None '''
        path = self.keyPath + name

        if os.path.isfile(path):
            with open(path, 'rb') as f:
                return f.read()

        return None

    #########################################################
    # Chutes
    #########################################################

    def chute(self, name):
        return None

    def chutes(self):
        return []

    def removeChute(self, name):
        pass

    def addChute(self, name):
        pass

    #########################################################
    # Network Resolution
    #########################################################

    def serverHost(self):
        ''' Different in production vs dev, etc. Returns hostname '''
        if self.mode == 'local':
            return 'localhost'

        return 'paradrop.io'

    def rifflePort(self):
        return _incrementingPort(8016, self.mode)

    def insecureRifflePort(self):
        ''' This isnt a thing yet. But it will be. '''
        return _incrementingPort(8017, self.mode)

    def webPort(self):
        return _incrementingPort(14321, self.mode)

    #########################################################
    # Path Resolution
    #########################################################

    def logPath(self):
        return self.logPath

    def keyPath(self):
        return self.keyPath

    def chutePath(self):
        return self.chutePath

    def uciPath(self):
        pass


#########################################################
# Utils
#########################################################

def _incrementingPort(base, mode):
    ''' Returns the given port plus some multiple of 10000 based on the passed mode '''

    if mode == 'production' or mode == 'local':
        return base

    if mode == 'development':
        return base + 10000

    if mode == 'test':
        return base + 10000 * 2

    return base


def createDefaultInfo(path):
    default = {
        'version': 1,
        'pdid': None,
        'chutes': [],
        'routers': [],
        'instances': []
    }

    writeYaml(default, path)


def validateInfo(contents):
    '''
    Error checking on the read YAML file. This is a temporary method.

    :param contents: the read-in yaml to check
    :type contents: dict.
    :returns: True if valid, else false
    '''
    INFO_REQUIRES = ['version', 'pdid', 'chutes', 'routers', 'instances']

    for k in INFO_REQUIRES:
        if k not in contents:
            return False

    # Check the validity of the contents

    return True


def writeYaml(contents, path):
    ''' Overwrites content with YAML representation at given path '''
    with open(path, 'w') as f:
        f.write(yaml.dump(contents, default_flow_style=False))


def loadYaml(path):
    ''' Return dict from YAML found at path '''
    with open(path, 'r') as f:
        return yaml.load(f.read())



'''
Wamp utility methods. 
'''

from autobahn.twisted.wamp import ApplicationSession, ApplicationRunner
from twisted.internet.defer import inlineCallbacks, returnValue, Deferred
from autobahn.wamp.types import RegisterOptions, SubscribeOptions, CallOptions, PublishOptions


from pdtools.lib.output import out


class BaseSession(ApplicationSession):

    """ Temporary base class for crossbar implementation """

    def __init__(self, config=None):
        ApplicationSession.__init__(self, config=config)
        self.pdid = config.extra

    @classmethod
    def start(klass, address, pdid, realm='crossbardemo', start_reactor=False, debug=False):
        '''
        Creates a new instance of this session and attaches it to the router
        at the given address and realm. The pdid is set manually now since we trust
        clients. Excessively.

        For now the realm is automatically set as a demo realm since we are not 
        using multiple realms.
        '''
        dee = Deferred()

        def maker(cfg):
            sess = klass(cfg)
            sess.dee = dee

            # print 'Assigning deferred to ' + sess
            return sess

        runner = ApplicationRunner(address, u'' + realm, extra=pdid, debug_wamp=debug, debug=debug,)
        runner.run(maker, start_reactor=start_reactor)

        return dee

    @inlineCallbacks
    def onJoin(self, details):
        print str(self.__class__.__name__) + ' crossbar session connected'
        yield

        # Inform whoever created us that the session has finished connecting.
        # Useful in situations where you need to fire off a single call and not a
        # full wamplet
        try:
            if self.dee is not None:
                yield self.dee.callback(self)
        except:
            # print 'No onJoin deferred callback set.'
            pass

    def pdRegister(self, name, *args, **kwargs):
        '''
        Very quick and dirty RPC registration that prepends the currently set PDID
        to the given method name.

        Incomplete. 
        '''

        if self.pdid is None:
            raise Exception("No PDID set! Cannot register!")

        ApplicationSession.register()

    # def onDisconnect(self):
    # print "disconnected"
    #     reactor.stop()

    ###################################################
    # Overridden CX interaction methods
    ###################################################

    '''
    Note: this first set of methods have all the REAL implemnetion of caller identification: 
    the caller's information is always passed along for every call. In the crossbar way of doing 
    things, however, whats passed along is a session id and not our pdid. 

    The second set of methods is temporary in that it manually passes the 
    current sessions' pdid. This is not secure, but will have to do for now in the 
    absence of crossbar router changes. 
    '''

    # def publish(self, topic, *args, **kwargs):
    #     kwargs['options'] = PublishOptions(disclose_me=True)
    #     return ApplicationSession.publish(self, topic, *args, **kwargs)

    # def subscribe(self, handler, topic=None, options=None):
    #     options = SubscribeOptions(details_arg='details')
    #     return ApplicationSession.subscribe(self, handler, topic=topic, options=options)

    # def call(self, procedure, *args, **kwargs):
    #     kwargs['options'] = CallOptions(disclose_me=True)
    #     return ApplicationSession.call(self, procedure, *args, **kwargs)

    # def register(self, endpoint, procedure=None, options=None):
    #     options = RegisterOptions(details_arg='details')
    #     return ApplicationSession.register(self, endpoint, procedure=procedure, options=options)

    def publish(self, pdid, topic, *args, **kwargs):
        # kwargs['options'] = PublishOptions(disclose_me=True)
        args = (self.pdid,) + args
        topic = _prepend(pdid, topic)
        # out.info('cxbr: (%s) publish (%s)' % (self.pdid, topic,))
        return ApplicationSession.publish(self, topic, *args, **kwargs)

    def subscribe(self, handler, pdid, topic=None, options=None):
        topic = _prepend(pdid, topic)
        out.info('cxbr: (%s) subscribe (%s)' % (self.pdid, topic,))
        return ApplicationSession.subscribe(self, handler, topic=topic, options=options)

    def call(self, pdid, procedure, *args, **kwargs):
        # kwargs['options'] = CallOptions(disclose_me=True)
        args = (self.pdid,) + args
        procedure = _prepend(pdid, procedure)
        # out.info('cxbr: (%s) calling (%s)' % (self.pdid, procedure,))
        return ApplicationSession.call(self, procedure, *args, **kwargs)

    def register(self, endpoint, procedure=None, options=None):
        # options = RegisterOptions(details_arg='session')
        procedure = _prepend(self.pdid, procedure)
        out.info('cxbr: (%s) register (%s)' % (self.pdid, procedure,))
        return ApplicationSession.register(self, endpoint, procedure=procedure, options=options)

    ###################################################
    # Access to the original methods, without convenience modifiers
    ###################################################
    def stockPublish(self, pdid, topic, *args, **kwargs):
        out.info('cxbr: (%s) publish (%s)' % (self.pdid, topic,))
        return ApplicationSession.publish(self, topic, *args, **kwargs)

    def stockSubscribe(self, handler, topic=None, options=None):
        out.info('cxbr: (%s) subscribe (%s)' % (self.pdid, topic,))
        return ApplicationSession.subscribe(self, handler, topic=topic, options=options)

    def stockCall(self, pdid, procedure, *args, **kwargs):
        out.info('cxbr: (%s) calling (%s)' % (self.pdid, procedure,))
        return ApplicationSession.call(self, procedure, *args, **kwargs)

    def stockRegister(self, endpoint, procedure=None, options=None):
        out.info('cxbr: (%s) registering (%s)' % (self.pdid, procedure,))
        return ApplicationSession.register(self, endpoint, procedure=procedure, options=options)


def _prepend(pdid, topic):
    '''
    In order to make subscription and execution code cleaner, this method automatically
    injects this classes pdid to the start of any publish or register call. 

    The topic is also converted to a unicode string. An underscore is inserted to the 
    start of every topic. No consideration is given to 'valid' topics-- thats on you.
    '''
    return u'' + pdid + '._' + topic


@inlineCallbacks
def cxCall(session, address, realm, extra=None):
    '''
    One shot crossbar utility method. Creates a session object and starts it,
    returns a deferred that contains the object. 

    Take a session and returns once the session has connected. 
    '''

    # WAMP ApplicationSession is already using 'd,' so went with something a little more creative
    dee = Deferred()

    def maker(cfg):
        sess = session(cfg)
        sess.dee = dee
        return sess

    runner = ApplicationRunner(address, realm, extra=extra, debug_wamp=False, debug=False,)
    yield runner.run(maker, start_reactor=False)

    session = yield dee
    returnValue(session)
